{---------------------------------------------------------------------------
|
|  Module: EnPrinter.Pas
|
|  Description: TPrinter like class which prints to Envision multi-page
|               graphic formats like PDF, TIF, and DCX.
|
|  History: Feb 14, 2004. MB First version
|
|---------------------------------------------------------------------------}
unit EnPrinter;

{$B-,X+}

interface

uses
    Classes,   { for TFileStream }
    EnMisc,    { for TImageFormat, GetTempFileName }
    EnDiGrph,  { for TDibGraphic }
    EnTifGr,   { for TTiffGraphic }
    Printers,  { for TPrinterOrientation  }
    Graphics;  { for TCanvas, TGraphic }

type

TFileExistAction = ( fxaOverwrite, fxaRaiseError );

TEnvisionPrinter = class(TObject)
protected

    FAborted         : Boolean;

    FPageNumber      : Integer;

    FPageHeight      : Integer;
    FPageWidth       : Integer;

    { MB Aug 25, 2006 }
    FXDotsPerInch    : Integer;
    FYDotsPerInch    : Integer;

    FImageFormat     : TImageFormat;

    FOutputFileName  : String;

    { Only not nil when printing. }
    FPageGraphic     : TTiffGraphic;

    FFileExistAction : TFileExistAction;

    FBufferFileName  : String;
    FBufferStream    : TFileStream;

    FIgnoreDefaultPrinter : Boolean;

    function GetPrinting : Boolean;

    function GetCanvas : TCanvas;

    procedure SetOutputFileName( const InOutputFileName : String );

    procedure CheckPrinting( const Active : Boolean );

    procedure SaveCurrentPage;
    function GetGraphic( const ImageNo     : Integer;
                         var   FreeGraphic : Boolean ) : TGraphic;

public
    constructor Create;
    destructor Destroy; override;

    property ImageFormat : TImageFormat read FImageFormat
                                        write FImageFormat;

    property Canvas : TCanvas read GetCanvas;

    property Aborted : Boolean read FAborted;

    property PageNumber : Integer read FPageNumber;

    { If manual changing the dimensions/dpi, IgnoreDefaultPrinter should
      be set true so that BeginDoc does not reset the settings based on the
      default printer }
    property PageHeight : Integer read FPageHeight write FPageHeight;
    property PageWidth : Integer read FPageWidth write FPageWidth;
    property XDotsPerInch : Integer read FXDotsPerInch write FXDotsPerInch;
    property YDotsPerInch : Integer read FYDotsPerInch write FYDotsPerInch;

    { MB Jun 15, 2010. If True, BeginDoc will not read settings from
      default printer }
    property IgnoreDefaultPrinter : Boolean read FIgnoreDefaultPrinter write FIgnoreDefaultPrinter;


    property Printing : Boolean read GetPrinting;

    procedure BeginDoc;
    procedure EndDoc;
    procedure Abort;

    procedure NewPage;

    property OutputFileName : String read FOutputFileName
                                     write SetOutputFileName;

    { default action when file exists is fxaOverwrite }
    property FileExistAction : TFileExistAction read FFileExistAction
                                                write FFileExistAction;

end;

function EnvisionPrinter : TEnvisionPrinter;

implementation

uses
    Forms,      { for Screen }
    Windows,    { for GetDeviceCaps }
    SysUtils;   { for Exception, DeleteFile }

var

FPrinter : TEnvisionPrinter;

{---------------------------------------------------------------------------}

function EnvisionPrinter : TEnvisionPrinter;
begin
    if FPrinter = nil then
        FPrinter := TEnvisionPrinter.Create;

    Result := FPrinter;
end;

{---------------------------------------------------------------------------}

constructor TEnvisionPrinter.Create;
begin
    inherited Create;

    FImageFormat          := ifBlackWhite;
    FFileExistAction      := fxaOverwrite;

    FIgnoreDefaultPrinter := False;

    { set default values of the page based on the currently selected
      printer }
    FXDotsPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
    FYDotsPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSY);

    FPageHeight   := Printer.PageHeight;
    FPageWidth    := Printer.PageWidth;
end;

{---------------------------------------------------------------------------}

destructor TEnvisionPrinter.Destroy;
begin
    FPageGraphic.Free;
    FPageGraphic := nil;

    FBufferStream.Free;
    FBufferStream := nil;

    inherited Destroy;
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.BeginDoc;
begin
    CheckPrinting(False);

    if (FileExistAction = fxaRaiseError) and
       FileExists( OutputFileName ) then
        raise Exception.Create('File already exists, ' + OutputFileName );

    FPageGraphic     := TTiffGraphic.Create;
    case FImageFormat of
        ifBlackWhite : FPageGraphic.Compression := tcGroup4;
        ifTrueColor  : FPageGraphic.Compression := tcJpeg;
        else
            FPageGraphic.Compression := tcZlib;
    end; { case }

    FBufferFileName := EnMisc.GetTempFileName( 'Prt', '.tmp' );
    FBufferStream   := TFileStream.Create( FBufferFileName, fmCreate );

    FAborted     := False;
    FPageNumber  := 1;

    if not FIgnoreDefaultPrinter then
    begin
        { MB Jan 12, 2009
          set default values of the page based on the currently selected
          printer on every call the BeginDoc, and the current printer
          may have changed orientation. }
        FXDotsPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
        FYDotsPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSY);

        FPageHeight  := Printer.PageHeight;
        FPageWidth   := Printer.PageWidth;
    end;

    FPageGraphic.NewImage( FPageWidth, FPageHeight, FImageFormat, nil,
                           FXDotsPerInch, FYDotsPerInch );

    FPageGraphic.Canvas.Font.PixelsPerInch := FYDotsPerInch;

    FPageGraphic.Fill(CWhiteRgb);
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.EndDoc;
var
    Graphic : TDibGraphic;
begin
    CheckPrinting(True);

    SaveCurrentPage;

    Graphic := NewDibGraphic(OutputFileName);
    BeginHourglass;
    try
        FBufferStream.Seek( 0, soFromBeginning );
        Graphic.MultiSaveToFile( OutputFileName, FPageNumber, GetGraphic );
    finally
        EndHourglass;
        Graphic.Free;
    end;

    FBufferStream.Free;
    FBufferStream := nil;

    DeleteFile(FBufferFileName);

    FPageGraphic.Free;
    FPageGraphic := nil;
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.Abort;
begin
    CheckPrinting(True);

    FAborted := True;

    FBufferStream.Free;
    FBufferStream := nil;

    DeleteFile(FBufferFileName);

    FPageGraphic.Free;
    FPageGraphic := nil;
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.NewPage;
begin
    CheckPrinting(True);

    SaveCurrentPage;

    Inc(FPageNumber);

    FPageGraphic.Fill(CWhiteRgb);
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.SaveCurrentPage;
var
    Stream : TMemoryStream;
    Size   : LongInt;
begin
    Stream := TMemoryStream.Create;
    try
        FPageGraphic.SaveToStream(Stream);
        Size := Stream.Size;

        FBufferStream.WriteBuffer( Size, SizeOf(Size) );
        Stream.Position := 0;
        {$WARNINGS OFF}
        FBufferStream.WriteBuffer( Stream.Memory^, Size );
        {$WARNINGS ON}
    finally
        Stream.Free;
    end;
end;

{---------------------------------------------------------------------------}

function TEnvisionPrinter.GetGraphic( const ImageNo     : Integer;
                                      var   FreeGraphic : Boolean ) : TGraphic;
var
    Size   : LongInt;
    Stream : TMemoryStream;
begin
    Result      := TTiffGraphic.Create;
    FreeGraphic := True;

    Stream := TMemoryStream.Create;
    try
        FBufferStream.ReadBuffer( Size, SizeOf(Size) );
        Stream.SetSize(Size);
        {$WARNINGS OFF}
        FBufferStream.ReadBuffer( Stream.Memory^, Size );
        {$WARNINGS ON}

        Stream.Position := 0;
        Result.LoadFromStream(Stream);
    finally
        Stream.Free;
    end;
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.CheckPrinting( const Active : Boolean );
begin
    if Active then
    begin
        if not Printing then
            raise Exception.Create( 'TEnvisionPrinter not printing');
    end
    else
    begin
        if Printing then
            raise Exception.Create( 'TEnvisionPrinter busy');
    end;
end;

{---------------------------------------------------------------------------}

function TEnvisionPrinter.GetPrinting : Boolean;
begin
    Result := (FPageGraphic <> nil);
end;

{---------------------------------------------------------------------------}

function TEnvisionPrinter.GetCanvas : TCanvas;
begin
    CheckPrinting(True);

    Result := FPageGraphic.Canvas;
end;

{---------------------------------------------------------------------------}

procedure TEnvisionPrinter.SetOutputFileName( const InOutputFileName : String );
begin
    CheckPrinting(False);

    FOutputFileName := InOutputFileName;
end;

{---------------------------------------------------------------------------}

initialization

    FPrinter := nil;

finalization

    FPrinter.Free;

end.
