{----------------------------------------------------------------------------
|
| Library: Envision
|
| Module: EnTrsOcr
|
| Description: Driver for Transym OCR engine.
|
| History: Mar 21, 2002. Michel Brazeau, first version
|
|---------------------------------------------------------------------------}
unit EnTrsOcr;

{$I Envision.Inc}

interface

uses
    Classes,  { for TStringList }
    TransOCR, { for TOCRJobInfo, ... }
    EnMisc,   { for TEnvisionProgressEvent }
    EnDiGrph, { for TDibGraphic }
    EnOcrEng; { for TOcrDriver }


type

TTransymOcrOptions = ( tooDeskew,
                       tooNoiseRemove,
                       tooLineRemove,
                       tooDeshade,
                       tooInvert,
                       tooSectioning,
                       tooMergeBreak,
                       tooLineReject,
                       tooCharacterReject,
                       tooUseDictionary );


TTransymOcrOptionsSet = set of TTransymOcrOptions;

TTransymOcrDriver = class(TOcrDriver)
protected
    FOptions  : TTransymOcrOptionsSet;

    { MB Apr 9, 2010. Modified so each driver instance gets a DLL handle
      so ocr can be done in multi threaded applications }
    FDLLHandle             : THandle;

    FOCRInitialise         : TOCRInitialise;
    FOCRShutdown           : TOCRShutdown;
    FOCRGetErrorMode       : TOCRGetErrorMode;
    FOCRSetErrorMode       : TOCRSetErrorMode;
    FOCRDoJob              : TOCRDoJob;
    FOCRWaitForJob         : TOCRWaitForJob;
    FOCRWaitForAnyJob      : TOCRWaitForAnyJob;
    FOCRGetJobDBInfo       : TOCRGetJobDBInfo;
    FOCRGetJobStatus       : TOCRGetJobStatus;
    FOCRGetJobStatusEx     : TOCRGetJobStatusEx;
    FOCRGetJobStatusMsg    : TOCRGetJobStatusMsg;
    FOCRGetNumPages        : TOCRGetNumPages;
    FOCRGetJobResults      : TOCRGetJobResults;
    FOCRGetLicenseInfo     : TOCRGetLicenceInfo;
    FOCRConvertTIFFToDIB   : TOCRConvertTIFFtoDIB;
    FOCRRotateMonoBitmap   : TOCRRotateMonoBitmap;

    procedure LoadTOCREngine;
    procedure UnloadTOCREngine;

    function GetEntryPoint( const Routine : AnsiString ) : Pointer;


public
    constructor Create; override;
    destructor Destroy; override;

    procedure Recognize( const Graphic       : TDibGraphic;
                         const Output        : TStrings;
                         const ProgressEvent : TEnvisionProgressEvent ); override;


    property Options : TTransymOcrOptionsSet read FOptions
                                             write FOptions;

end;

{--------------------------------------------------------------------------}

implementation

uses
    SyncObjs, { for TCritialSection }
    Windows,  { for THandle }
    SysUtils, { for Exception }
    EnBmpGr,  { for TBitmapGraphic }
    ClipBrd;  { for TClipBoard }

const

CTOCRDllFileName = 'TOCRDLL.DLL';

var

OcrCS                        : TCriticalSection;

{--------------------------------------------------------------------------}

procedure SetJobOptions( const Options    : TTransymOcrOptionsSet;
                         var   JobInfo    : TOCRJobInfo );
begin
    FillChar( JobInfo, SizeOf(JobInfo), 0);
    JobInfo.JobType   := TOCRJOBTYPE_DIBFILE;
    JobInfo.PageNo    := 0;

    JobInfo.ProcessOptions.InvertWholePage     := OCRFalse;
    JobInfo.ProcessOptions.Orientation         := TOCRJOBORIENT_OFF;

    if tooDeskew in Options then
        JobInfo.ProcessOptions.DeskewOff  := OCRFalse
    else
        JobInfo.ProcessOptions.DeskewOff  := OCRTrue;

    if tooNoiseRemove in Options then
        JobInfo.ProcessOptions.NoiseRemoveOff  := OCRFalse
    else
        JobInfo.ProcessOptions.NoiseRemoveOff  := OCRTrue;

    if tooLineRemove in Options then
        JobInfo.ProcessOptions.LineRemoveOff  := OCRFalse
    else
        JobInfo.ProcessOptions.LineRemoveOff  := OCRTrue;

    if tooDeshade in Options then
        JobInfo.ProcessOptions.DeshadeOff  := OCRFalse
    else
        JobInfo.ProcessOptions.DeshadeOff  := OCRTrue;

    if tooInvert in Options then
        JobInfo.ProcessOptions.InvertOff  := OCRFalse
    else
        JobInfo.ProcessOptions.InvertOff  := OCRTrue;

    if tooSectioning in Options then
        JobInfo.ProcessOptions.SectioningOn  := OCRTrue
    else
        JobInfo.ProcessOptions.SectioningOn  := OCRFalse;

    if tooMergeBreak in Options then
        JobInfo.ProcessOptions.MergeBreakOff  := OCRFalse
    else
        JobInfo.ProcessOptions.MergeBreakOff  := OCRTrue;

    if tooLineReject in Options then
        JobInfo.ProcessOptions.LineRejectOff  := OCRFalse
    else
        JobInfo.ProcessOptions.LineRejectOff  := OCRTrue;

    if tooCharacterReject in Options then
        JobInfo.ProcessOptions.CharacterRejectOff  := OCRFalse
    else
        JobInfo.ProcessOptions.CharacterRejectOff  := OCRTrue;

    if tooUseDictionary in Options then
        JobInfo.ProcessOptions.LexOff  := OCRFalse
    else
        JobInfo.ProcessOptions.LexOff  := OCRTrue;
end;

{--------------------------------------------------------------------------}

constructor TTransymOcrDriver.Create;
begin
    inherited Create;
    
    LoadTOCREngine;

    { set the default options }
    FOptions := [ tooDeskew,
                  tooNoiseRemove,
                  tooLineRemove,
                  tooDeshade,
                  tooSectioning,
                  tooMergeBreak,
                  tooLineReject,
                  tooUseDictionary,
                  tooCharacterReject
                ];

end;

{--------------------------------------------------------------------------}

destructor TTransymOcrDriver.Destroy;
begin
    UnloadTOCREngine;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure TTransymOcrDriver.Recognize( const Graphic       : TDibGraphic;
                                       const Output        : TStrings;
                                       const ProgressEvent : TEnvisionProgressEvent );
var
    JobStatus          : LongInt;
    ResultSize         : LongInt;
    pOCRResults        : TpOCRResults;
    {$Warnings Off}
    ResultsPtr         : Pointer;
    {$Warnings On}
    ItemNo             : LongInt;
    Text               : String;
    JobInfo            : TOCRJobInfo;
    LastPercent        : Byte;
    Progress           : Single;
    Orientation        : LongInt;
    JobNo              : LongInt;

    FileName           : AnsiString;
    BitmapGraphic      : TBitmapGraphic;

begin
    { MB Nov 22, 2002
      workaround a bug where somethings the first line of text was not recognized.
      Just add one black pixel in the top/left corner. }
    Graphic.RGB[0,0] := MakeRgb(0,0,0);

    { MB Apr 8, 2010. Modified to use temporary file instead of clipboard for
      multi threaded support. Use random prefix to ensure don't get
      name clashes }
    FileName := '';
    try
        FileName := EnMisc.GetTempFileName( 'en' + AnsiString(IntToStr(Random(999999))), 'tmp' );
        { create file immediately to prevent conflicts }
        FileClose(FileCreate(FileName));

        BitmapGraphic := TBitmapGraphic.Create;
        try
            BitmapGraphic.Assign(Graphic);
            BitmapGraphic.SaveToFile(FileName);
        finally
            BitmapGraphic.Free;
        end;

        { MB Apr 9, 2010. Transym engine is not thread safe. Use critical
          section }
        EnTrsOcr.OcrCS.Enter;
        try
            JobNo    := -1;

            SetJobOptions( FOptions, JobInfo );

            JobInfo.InputFile := PAnsiChar(FileName);

            {$Warnings Off}
            CheckOCRCall(FOCRInitialise(@JobNo));
            {$Warnings On}


            {$Warnings Off}
            CheckOCRCall(FOCRDoJob(JobNo, @JobInfo));
            {$Warnings On}

            JobStatus := -1;

            LastPercent := DoProgress( Self, ProgressEvent, 0, 0, 100, -1 );
            repeat
                {$Warnings Off}
                CheckOCRCall(FOCRGetJobStatusEx(JobNo, @JobStatus, @Progress, @Orientation));
                {$Warnings On}

                LastPercent := DoProgress( Self, ProgressEvent, SafeTrunc(Progress*100), 0, 100, LastPercent );

                if JobStatus = TOCRJOBSTATUS_BUSY then
                    Sleep(0);

            until (JobStatus <> TOCRJOBSTATUS_BUSY);

            DoProgress( Self, ProgressEvent, 100, 0, 100, LastPercent );


            if JobStatus = TOCRJOBSTATUS_DONE then
            begin
                ResultSize := 0;

                {$Warnings Off}
                CheckOCRCall( FOCRGetJobResults( JobNo, @ResultSize, nil ));
                {$Warnings On}

                if ResultSize > 0 then
                begin
                    {$Warnings Off}
                    GetMem( ResultsPtr, ResultSize );
                    {$Warnings On}
                    try
                        {$Warnings Off}
                        pOCRResults := ResultsPtr;

                        pOCRResults^.Header.NumItems := 0;

                        CheckOCRCall( FOCRGetJobResults( JobNo, @ResultSize, ResultsPtr ));

                        Text := '';
                        for ItemNo := 0 to pOCRResults^.Header.NumItems-1 do
                            Text := Text + Char( pOCRResults.Items[ItemNo].OCRCha);

                        { Nov 14, 2010. Typecast to PWideChar. Thanks to Mario Vermeulen }
                        {$ifdef __Delphi2009AndAbove}
                        Output.SetText(PWideChar(Text));
                        {$else}
                        Output.SetText(PChar(Text));
                        {$endif}

                        {$Warnings On}

                    finally
                        {$Warnings Off}
                        FreeMem( ResultsPtr );
                        {$Warnings On}
                    end;
                end;
            end;

            if JobNo <> -1 then
                CheckOCRCall(FOCRShutDown(JobNo));

        finally
                EnTrsOcr.OcrCS.Leave;
        end;

    finally
        if FileName <> '' then
            DeleteFile(FileName);
    end;
end;

{---------------------------------------------------------------------------}

{$Warnings Off}
function TTransymOcrDriver.GetEntryPoint( const Routine : AnsiString ) : Pointer;
begin
    Result := GetProcAddress( FDllHandle, PAnsiChar( Routine ));

    if Result = nil then
         raise Exception.Create( 'Routine ' + Routine + ' not found in ' + CTOCRDllFileName );
end;
{$Warnings On}

{---------------------------------------------------------------------------}

procedure TTransymOcrDriver.LoadTOCREngine;
begin
    if FDLLHandle <> 0 then
        raise Exception.Create('TOCR DLL already loaded');

    {$Warnings Off}
    FDLLHandle := LoadLibrary(PChar(CTOCRDllFileName));
    {$Warnings On}

    if FDLLHandle = 0 then
        raise Exception.Create( 'Unable to load ' + CTOCRDllFileName + '.'#13#13 +
                                'Make sure Transym OCR is installed.'#13 +
                                'A trial version may be downloaded from:'#13#13 +
                                'www.transym.com '#13 );

    FOCRInitialise         :=  GetEntryPoint( 'TOCRInitialise' );
    FOCRShutdown           :=  GetEntryPoint( 'TOCRShutdown' );
    FOCRGetErrorMode       :=  GetEntryPoint( 'TOCRGetErrorMode' );
    FOCRSetErrorMode       :=  GetEntryPoint( 'TOCRSetErrorMode' );
    FOCRDoJob              :=  GetEntryPoint( 'TOCRDoJob' );
    FOCRWaitForJob         :=  GetEntryPoint( 'TOCRWaitForJob' );
    FOCRWaitForAnyJob      :=  GetEntryPoint( 'TOCRWaitForAnyJob' );
    FOCRGetJobDBInfo       :=  GetEntryPoint( 'TOCRGetJobDBInfo' );
    FOCRGetJobStatus       :=  GetEntryPoint( 'TOCRGetJobStatus' );
    FOCRGetJobStatusEx     :=  GetEntryPoint( 'TOCRGetJobStatusEx' );
    FOCRGetJobStatusMsg    :=  GetEntryPoint( 'TOCRGetJobStatusMsg' );
    FOCRGetNumPages        :=  GetEntryPoint( 'TOCRGetNumPages' );
    FOCRGetJobResults      :=  GetEntryPoint( 'TOCRGetJobResults' );
    FOCRGetLicenseInfo     :=  GetEntryPoint( 'TOCRGetLicenceInfo' );
    FOCRConvertTIFFToDIB   :=  GetEntryPoint( 'TOCRConvertTIFFtoDIB' );
    FOCRRotateMonoBitmap   :=  GetEntryPoint( 'TOCRRotateMonoBitmap' );
end;

{--------------------------------------------------------------------------}

procedure TTransymOcrDriver.UnloadTOCREngine;
begin
    if FDLLHandle <> 0 then
    begin
        FreeLibrary(FDLLHandle);
        FDLLHandle := 0;
    end;
end;

{--------------------------------------------------------------------------}

initialization

    EnTrsOcr.OcrCS    := TCriticalSection.Create;

finalization

    EnTrsOcr.OcrCS.Free;

end.

